from satispy import Variable, Cnf
from satispy.solver import Minisat

v1 = Variable('v1')
v2 = Variable('v2')
v3 = Variable('v3')

exp = v1 & v2 | v3

solver = Minisat()

solution = solver.solve(exp)

if solution.success:
    print "Found a solution:"
    print v1, solution[v1]
    print v2, solution[v2]
    print v3, solution[v3]
else:
    print "The expression cannot be satisfied"



import numpy

class Variable(object):
    def __init__(self, name, inverted=False):
        self.name = name
        self.inverted = inverted

    def __neg__(self):
        v = Variable(self.name)
        v.inverted = not self.inverted
        return v

    def __and__(self, other):
        c = Cnf.create_from(self)
        return c & other

    def __or__(self, other):
        c = Cnf.create_from(self)
        return c | other

    def __xor__(self, other):
        c = Cnf.create_from(self)
        return c ^ other

    def __rshift__(self, other): # implies
        c = Cnf.create_from(self)
        return -c | other

    def __str__(self):
        return ("-" if self.inverted else "") + self.name

    def __eq__(self, other):
        return self.name == other.name and self.inverted == other.inverted

    def __hash__(self):
        return hash(self.name) ^ hash(self.inverted)

    def __cmp__(self, other):
        if self == other:
            return 0
        if (self.name, self.inverted) < (other.name, other.inverted):
            return -1
        else:
            return 1

class Cnf(object):
    def __init__(self):
        self.dis = []

    @classmethod
    def create_from(cls, x):
        if isinstance(x, Variable):
            cnf = Cnf()
            cnf.dis = [frozenset([x])]
            return cnf
        elif isinstance(x, cls):
            return x
        else:
            raise Exception("Could not create a Cnf object from %s" % str(type(x)))

    def __and__(self, other):
        other = Cnf.create_from(other)
        result = Cnf()
        result.dis = self.dis + other.dis
        return result

    def __or__(self, other):
        other = Cnf.create_from(other)

        if len(self.dis) > 0 and len(other.dis) > 0:
            new_dis = []
            for d1, d2 in [(d1,d2) for d1 in self.dis for d2 in other.dis]:
                d3 = d1 | d2
                new_dis.append(d3)
        elif len(self.dis) == 0:
            new_dis = other.dis
        else:
            new_dis = self.dis

        c = Cnf()
        c.dis = new_dis
        return c

    def __xor__(self, other):
        return (self | other) & (-self | -other)

    def __neg__(self):
        cnfs = []

        for d in self.dis:
            c = Cnf()
            for v in d:
                c.dis.append(frozenset([-v]))
            cnfs.append(c)

        ret = cnfs.pop()
        for cnf in cnfs:
            ret |= cnf

        return ret

    def __rshift__(self, other): # implies
        return -self | other

    def __str__(self):
        ret = []
        for d in self.dis:
            ret.append(" | ".join(map(str,d)))
        return "(" + ") & (".join(ret) + ")"

    def __eq__(self, other):
        return self.dis == other.dis

    def __hash__(self):
        return hash(self.dis)

class BitVariable(object):
    varmap = {}
    def __init__(self, name, varz):
        if BitVariable.varmap.has_key(name):
            self.varno = BitVariable.varmap[name]
        else:
            self.varno = len(BitVariable.varmap)
            BitVariable.varmap[name] = self.varno
        self.name = name
        self.inverted = 0
        self.chunk = self.varno / 32
        #print (((self.varno % 16) * 2) + self.inverted)
        self.bitfield = (1 << (((self.varno % 31) * 2) + self.inverted))
        self.varz = varz

    def __neg__(self):
        v = BitVariable(self.name, self.varz)
        v.inverted = 1 - self.inverted
        v.bitfield = (1 << (((v.varno % 31) * 2) + v.inverted))
        return v

    def __and__(self, other):
        c = BitCnf.create_from(self, self.varz)
        return c & other

    def __or__(self, other):
        c = BitCnf.create_from(self, self.varz)
        return c | other

    def __xor__(self, other):
        c = BitCnf.create_from(self, self.varz)
        return c ^ other

    def __rshift__(self, other): # implies
        c = BitCnf.create_from(self, self.varz)
        return -c | other

    def __str__(self):
        return ("-" if self.inverted else "") + self.name

    def __eq__(self, other):
        return self.name == other.name and self.inverted == other.inverted

    def __hash__(self):
        return hash(self.name) ^ hash(self.inverted)

    def __cmp__(self, other):
        if self == other:
            return 0
        if (self.name, self.inverted) < (other.name, other.inverted):
            return -1
        else:
            return 1

class BitCnf(object):
    def __init__(self, varz):
        self.clauses = []   # Bit table
        self.varz = varz    # Number of possible variables

    @classmethod
    def _new_clause(cls,varz):
        return numpy.zeros(varz/32+1, dtype=numpy.int64)

    def dedup_clauses(self):
        """Eliminate double clauses
        """
        newclauses = []
        for clause in self.clauses:
            if clause not in newclauses:
                newclauses.append(clause)
        self.clauses = newclauses

    def dedup_vars(self):
        """Eliminates e | -e type of expressions
        """
        lsmask = int('10'*32,2)
        rsmask = int('01'*32,2)

        for clause in self.clauses:
            numpy.vectorize(lambda a: a & ~(a & (a << 1) & lsmask) & ~(a & (a >> 1) & rsmask))(self.clauses)

    @classmethod
    def create_from(self,v,varz):
        if isinstance(v, BitVariable):
            cnf = BitCnf(varz)
            c = BitCnf._new_clause(varz)
            c[v.chunk] |= v.bitfield
            cnf.clauses.append(c)
            return cnf
        else:
            return v

    def __iand__(self, other):
        other = self.create_from(other, self.varz)

        for clause in other.clauses:
            self.clauses.append(clause[:])
        return self

    def __or__(self, other):
        other = self.create_from(other, self.varz)
        newcnf = BitCnf(self.varz)

        if len(self.clauses) > 0 and len(other.clauses) > 0:
            for c1, c2 in [(c1,c2) for c1 in self.clauses for c2 in other.clauses]:
                newcnf.clauses.append(c1 | c2) # Numpy for the rescue
        elif len(self.clauses) == 0:
            for clause in other.clauses:
                newcnf.clauses.append(clause[:])
        else:
            for clause in self.clauses:
                newcnf.clauses.append(clause[:])

        return newcnf

    def __and__(self, other):
        other = self.create_from(other, self.varz)
        newcnf = BitCnf(self.varz)
        for clause in self.claues:
            newcnf.clauses.append(clause[:])
        if isinstance(other, BitVariable):
            c = BitCnf._new_clause(self.varz)
            c[other.chunk] |= other.bitfield
            newcnf.clauses.append(c)
            return newcnf
        else:
            for clause in other.claues:
                newcnf.clauses.append(clause[:])
            return newcnf



# Naive implementation of Hamilton-cycle searching in a graph
#
# Based on the following paper:
#
# Formalizing Dangerous SAT Encodings
# Alexander Hertel, Philipp Hertel, and Alasdair Urquhart
# Department of Computer Science,
# University of Toronto, Toronto ON M5S 3G4, Canada,
# [ahertel][philipp][urquhart]@cs.toronto.edu

from satispy import Variable, Cnf
from satispy.solver import Minisat

import networkx as nx

N = 6
M = 6

g = nx.grid_2d_graph(N, M)

lennodes = len(g.nodes())
gnodes = g.nodes()

exp = Cnf()

# Variables
# We create a variable for every node that
# marks the position of that node in the path
print "Creating variables..."
varz = {}
for n in gnodes:
    varz[n] = []
    for i in xrange(lennodes):
        # n, m, position
        varz[n].append(Variable("%d_%d_%d" % (n[0], n[1], i)))

# Total (X)
print "Creating total clauses..."
for i in gnodes:
    c = Cnf()
    for j in xrange(lennodes):
        c |= varz[i][j]
    exp &= c

# Onto
print "Creating onto clauses..."
for j in xrange(lennodes):
    c = Cnf()
    for i in gnodes:
        c |= varz[i][j]
    exp &= c

# 1-1 (X)
print "Creating 1-1 calues..."
for j in xrange(lennodes):
    print j
    for i1 in gnodes:
        for i2 in gnodes:
            if i1 != i2:
                exp &= -varz[i1][j] | -varz[i2][j]

# Fn
print "Creating Fn calues..."
for i in xrange(lennodes):
    print i
    for j1 in gnodes:
        for j2 in gnodes:
            if i1 != i2:
                exp &= -varz[i][j1] | -varz[i][j2]

# Edge
print "Adding edge clauses..."
for j in xrange(lennodes):
    print j
    for i in gnodes:
        for k in gnodes:
            if i != k and k not in g.neighbors(i):
                exp &= -varz[i][j] | -varz[k][(j+1) % lennodes]
            
# Enabling minisat to write to stdout
solver = Minisat('minisat %s %s')

print "Solving..."
solution = solver.solve(exp)

if not solution.success:
    print "There is no Hamilton cycle in the graph."
    exit()

print "Extracting solution..."
path = []
for n in g.nodes():
    nodepos = -1
    for j in xrange(lennodes):
        if solution[varz[n][j]]:
            nodepos = j
            break
    path.append((nodepos, n))

path.sort()

for n in path:
    print n[0], n[1]



#!/usr/bin/env python
# -*- coding: utf8 -*-

from distutils.core import setup, Extension

cnfmodule = Extension(
    'satispy.cnf',
    sources = [
        'src/ext/cnf_module.c',
        'src/ext/cnf_cnf.c',
        'src/ext/cnf_variable.c'
    ]
)

setup(
    name='satispy',
    version='1.0a4',
    description='An interface to SAT solver tools (like minisat)',
    author='FÁBIÁN Tamás László',
    author_email='giganetom@gmail.com',
    url='https://github.com/netom/satispy/',
    download_url='https://github.com/netom/satispy/tarball/1.0a4#egg=satispy-1.0a4',
    license='BSD License',
    platforms='OS Independent',
    package_dir = {'': 'src/python'},
    packages=['satispy', 'satispy.io', 'satispy.solver'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Plugins',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Software Development :: Libraries'
    ],
    ext_modules = [cnfmodule]
)



#!/usr/bin/env python

import os
import sys
import copy
import unittest

from satispy import *
from satispy.io import *
from satispy.solver import *

class VariableTest(unittest.TestCase):
    def testVariable(self):
        v = Variable("name")
        self.assertEqual("name", str(v))
        self.assertEqual("-name", str(-v))

    def testName(self):
        v1 = Variable('v1')
        self.assertEqual('v1', v1.name)

    def testInverted(self):
        v1 = Variable('v1')
        v2 = Variable('v2', 1)

        self.assertEqual(0, v1.inverted)
        self.assertEqual(1, v2.inverted)
        self.assertRaises(ValueError, Variable, 'v3', 2)

    def testNegation(self):
        v1 = Variable('v1')
        v2 = Variable('v2', 1)

        self.assertEqual(1, (-v1).inverted)
        self.assertEqual(0, (-v2).inverted)

class CnfTest(unittest.TestCase):
    def testInit(self):
        cnf = Cnf()
        self.assertEqual("", cnf.getBuffer())

    #~ def testAnd(self):
        #~ v1 = Variable("v1")
        #~ v2 = Variable("v2")
        #~ cnf = v1 & v2
        #~ self.assertEqual(set([frozenset([v1]),frozenset([v2])]), set(cnf.dis))

    def testOr(self):
        v1 = Variable("v1")
        v2 = Variable("v2")

        cnf = v1 | v2

        val = 1 << v1.number | 1 << v2.number

        self.assertEqual(val, reduce(lambda x, y: 256*x+ord(y), cnf.getBuffer(), 0))
#~ 
        #~ cnf2 = cnf1 | v3
#~ 
        #~ self.assertEqual(set([frozenset([v1,v2,v3])]), set(cnf2.dis))
#~ 
        #~ # Test empty CNF or
        #~ cnf = Cnf()
        #~ cnf |= v1
#~ 
        #~ self.assertEqual(set([frozenset([v1])]), set(cnf.dis))
#~ 
    #~ def testCreateFrom(self):
        #~ v1 = Variable("v1")
        #~ self.assertEqual(set([frozenset([v1])]), set(Cnf.create_from(v1).dis))
#~ 
    #~ def testMixed(self):
        #~ v1 = Variable("v1")
        #~ v2 = Variable("v2")
        #~ v3 = Variable("v3")
        #~ v4 = Variable("v4")
#~ 
        #~ self.assertEqual(
            #~ set([
                #~ frozenset([v1,v2]),
                #~ frozenset([v3])
            #~ ]),
            #~ set(((v1 | v2) & v3).dis)
        #~ )
#~ 
        #~ # Distribution
        #~ self.assertEqual(
            #~ set([
                #~ frozenset([v1,v3]),
                #~ frozenset([v2,v3])
            #~ ]),
            #~ set(((v1 & v2) | v3).dis)
        #~ )
#~ 
        #~ # Double distribution
        #~ self.assertEqual(
            #~ set([
                #~ frozenset([v1,v3]),
                #~ frozenset([v1,v4]),
                #~ frozenset([v2,v3]),
                #~ frozenset([v2,v4])
            #~ ]),
            #~ set(((v1 & v2) | (v3 & v4)).dis)
        #~ )
#~ 
    #~ def testNegation(self):
        #~ v1 = Variable("v1")
        #~ v2 = Variable("v2")
        #~ v3 = Variable("v3")
#~ 
        #~ self.assertEqual(
            #~ set([frozenset([-v1])]),
            #~ set((-Cnf.create_from(v1)).dis)
        #~ )
#~ 
    #~ def testXor(self):
        #~ v1 = Variable("v1")
        #~ v2 = Variable("v2")
#~ 
        #~ cnf = v1 ^ v2
#~ 
        #~ self.assertEqual(
            #~ set([
                #~ frozenset([v1,v2]),
                #~ frozenset([-v1,-v2])
            #~ ]),
            #~ set(cnf.dis)
        #~ )
#~ 
    #~ def testImplication(self):
        #~ v1 = Variable("v1")
        #~ v2 = Variable("v2")
#~ 
        #~ self.assertEqual(
            #~ set([frozenset([-v1,v2])]),
            #~ set((v1 >> v2).dis)
        #~ )
#~ 
#~ class DimacsTest(unittest.TestCase):
    #~ def testDimacs(self):
        #~ v1 = Variable("v1")
        #~ v2 = Variable("v2")
        #~ v3 = Variable("v3")
        #~ v4 = Variable("v4")
#~ 
        #~ cnf1 = (v1 | v2) & (-v3 | v4) & (-v1 | v3 | -v4)
#~ 
        #~ io = DimacsCnf()
#~ 
        #~ s = io.tostring(cnf1)
#~ 
        #~ # Must be true because of Variable ordering, and
        #~ # the guarantee of numbering during IO
        #~ self.assertEqual(io.varname(v1), '1')
        #~ self.assertEqual(io.varname(v2), '2')
        #~ self.assertEqual(io.varname(v3), '3')
        #~ self.assertEqual(io.varname(v4), '4')
        #~ self.assertEqual(io.varobj('1'), v1)
        #~ self.assertEqual(io.varobj('2'), v2)
        #~ self.assertEqual(io.varobj('3'), v3)
        #~ self.assertEqual(io.varobj('4'), v4)
#~ 
        #~ cnf2 = io.fromstring(s)
#~ 
        #~ # Must be true due to io conventions
        #~ self.assertEqual(io.varname(v1), '1')
        #~ self.assertEqual(io.varname(v2), '2')
        #~ self.assertEqual(io.varname(v3), '3')
        #~ self.assertEqual(io.varname(v4), '4')
        #~ self.assertEqual(io.varobj('1'), v1)
        #~ self.assertEqual(io.varobj('2'), v2)
        #~ self.assertEqual(io.varobj('3'), v3)
        #~ self.assertEqual(io.varobj('4'), v4)
#~ 
        #~ # The frozenset is deterministic, so this will be ture
        #~ # if everything else is OK
        #~ self.assertEqual(cnf1, cnf2)
#~ 
#~ class MinisatTest(unittest.TestCase):
#~ 
    #~ def testSolution(self):
        #~ v1 = Variable("v1")
        #~ v2 = Variable("v2")
        #~ v3 = Variable("v3")
        #~ v4 = Variable("v4")
#~ 
        #~ cnf = (v1 | v2) & (-v3 | v4) & (-v1 | v3 | -v4)
        #~ #  v1, -v2, -v3, -v4
        #~ #  v1, -v2,  v3,  v4
        #~ #  v1,  v2, -v3, -v4
        #~ #  v1,  v2,  v3,  v4
        #~ # -v1,  v2, -v3, -v4
        #~ # -v1,  v2, -v3,  v4
        #~ # -v1,  v2,  v3,  v4
#~ 
        #~ solver = Minisat()
#~ 
        #~ sol = solver.solve(cnf)
#~ 
        #~ self.assertTrue(
            #~ (sol[v1],sol[v2], sol[v3], sol[v4])
            #~ in
            #~ [
                #~ (True,  False, False, False),
                #~ (True,  False, True,  True ),
                #~ (True,  True,  False, False),
                #~ (True,  True,  True,  True ),
                #~ (False, True,  False, False),
                #~ (False, True,  False, True ),
                #~ (False, True,  True,  True )
            #~ ]
        #~ )


if __name__ == "__main__":
    unittest.main()



from satispy.io import DimacsCnf
from satispy import Variable

from subprocess import call
from tempfile import NamedTemporaryFile

class Solution(object):
    def __init__(self, success=False, varmap={}):
        self.success = success
        self.varmap = varmap

    def __getitem__(self, i):
        return self.varmap[i]

class Minisat(object):
    COMMAND = 'minisat %s %s > /dev/null'

    def __init__(self, command=COMMAND):
        self.command = command

    def solve(self, cnf):
        s = Solution()

        infile = NamedTemporaryFile(mode='w')
        outfile = NamedTemporaryFile(mode='r')

        io = DimacsCnf()
        infile.write(io.tostring(cnf))
        infile.flush()

        ret = call(self.command % (infile.name, outfile.name), shell=True)

        infile.close()

        if ret != 10:
            return s

        s.success = True

        lines = outfile.readlines()[1:]

        for line in lines:
            varz = line.split(" ")[:-1]
            for v in varz:
                v = v.strip()
                value = v[0] != '-'
                v = v.lstrip('-')
                vo = io.varobj(v)
                s.varmap[vo] = value

        # Close deletes the tmp files
        outfile.close()

        return s
